#!/bin/bash
TOOLS_VERSION=`cat changelog debian/changelog 2>/dev/null | head -qn 1 `
TOOLS_VERSION="${TOOLS_VERSION%%)*}"
TOOLS_VERSION="${TOOLS_VERSION##*(}"
echo "VERSION = ${TOOLS_VERSION%%-*}" >libguytools_version.pro.inc

